from .stopbyacc import BStopByAcc
from .stopbyaccdelta import BStopByAccDelta
from .stopbyloss import BStopByLoss
from .stopbylossdelta import BStopByLossDelta
from .stopbyoverfitting import BStopByOverfitting
from .stopbybyzh import BStopByBYZH

from .reloadbyloss import BReloadByLoss

__all__ = [
    'BStopByAcc',
    'BStopByAccDelta',
    'BStopByLoss',
    'BStopByLossDelta',
    'BStopByOverfitting',
    'BStopByBYZH',

    'BReloadByLoss',
]


from ..drawer.terminal import BColor
class BConfig():
    def __init__(self):
        self.dict = dict()

    # set-方法
    def set(self, key, value):
        if not isinstance(key, str):
            raise Exception(f"{BColor.YELLOW}key({str(key)}) must be str{BColor.RESET}")

        self.dict[key] = value

    # show-方法
    def show_all(self):
        print(self)

    # get-方法
    def get_str(self, key):
        self.__check(key)
        return str(self.dict[key])
    def get_int(self, key):
        self.__check(key)
        return int(self.dict[key])
    def get_float(self, key):
        self.__check(key)
        return float(self.dict[key])
    def get_bool(self, key):
        '''
        只有value为["False", "0", "None"]时，返回False
        '''
        self.__check(key)

        if self.get_str(key) in ["False", "0", "None"]:
            return False
        elif self.get_str(key) in ["True", "1"]:
            return True
        else:
            raise Exception(f"{BColor.YELLOW}value({str(key)}) cannot change to bool{BColor.RESET}")

    # save-方法
    def to_pickle(self, path):
        import pickle
        with open(path, 'wb') as f:
            pickle.dump(self.dict, f)
    def from_pickle(self, path):
        import pickle
        with open(path, 'rb') as f:
            self.dict = pickle.load(f)
    def to_json(self, path):
        import json
        with open(path, 'w') as f:
            json.dump(self.dict, f)

    def from_json(self, path):
        import json
        with open(path, 'r') as f:
            self.dict = json.load(f)

    def to_yaml(self, path):
        import yaml
        with open(path, 'w') as f:
            yaml.dump(self.dict, f)
    def from_yaml(self, path):
        import yaml
        with open(path, 'r') as f:
            self.dict = yaml.load(f, Loader=yaml.FullLoader)
    def to_ini(self, path):
        import configparser
        config = configparser.ConfigParser()
        config['config'] = self.dict
        with open(path, 'w') as f:
            config.write(f)

    def from_ini(self, path):
        import configparser
        config = configparser.ConfigParser()
        config.read(path)
        self.dict = dict(config['config'])

    def to_csv(self, path):
        import csv
        with open(path, 'w', newline='') as f:
            writer = csv.writer(f)
            for key, value in self.dict.items():
                writer.writerow([key, value])

    def from_csv(self, path):
        import csv
        with open(path, 'r') as f:
            reader = csv.reader(f)
            self.dict = {rows[0]: rows[1] for rows in reader}

    # 工具-方法
    def __str__(self):
        result = "default:\n" + "\n".join([f"\t({key} -> {value})" for key, value in self.dict.items()])
        return result
    def __check(self, key):
        # 检查key是否是字符串
        if not isinstance(key, str):
            raise Exception(f"{BColor.YELLOW}key({str(key)}) must be str{BColor.RESET}")



if __name__ == '__main__':
    a = BConfig()

    a.set('a', 'None')
    a.set('b', '123')
    a.set('345', '33333')
    a.set('a532', 'No32ne')
    a.set('b13', '123412')
    a.set('321345', '33342333')

    a.to_csv('config.csv')
    a.from_csv('config.csv')
    a.to_yaml('config.yaml')
    a.from_yaml('config.yaml')
    a.to_pickle('config.pkl')
    a.from_pickle('config.pkl')
    a.to_json('config.json')
    a.from_json('config.json')

    print(a.get_str('b13'))
    print(a)


import torch.nn as nn
import torch

class BLenet5(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        # [b, 6, 32, 32]
        self.layer1 = nn.Sequential(
            nn.Conv2d(1, 6, kernel_size=5, stride=1, padding=0),
            nn.BatchNorm2d(6),
            nn.ReLU(),
        )
        # [b, 6, 28, 28]
        self.subsample1 = nn.MaxPool2d(kernel_size=2, stride=2)  # 下采样
        # [b, 6, 14, 14]
        self.layer2 = nn.Sequential(
            nn.Conv2d(6, 16, kernel_size=5, stride=1, padding=0),
            nn.BatchNorm2d(16),
            nn.ReLU(),
        )
        # [b, 16, 10, 10]
        self.subsample2 = nn.MaxPool2d(kernel_size=2, stride=2)  # 下采样
        # [b, 16, 5, 5]

        # 全连接
        self.L1 = nn.Linear(400, 120)
        self.relu = nn.ReLU()
        self.L2 = nn.Linear(120, 84)
        self.relu1 = nn.ReLU()
        self.L3 = nn.Linear(84, num_classes)

    def forward(self, x):
        x = self.layer1(x)
        x = self.subsample1(x)
        x = self.layer2(x)
        x = self.subsample2(x)
        # 将上一步输出的16个5×5特征图中的400个像素展平成一维向量，以便下一步全连接
        x = x.reshape(x.size(0), -1)
        # 全连接
        x = self.L1(x)
        x = self.relu(x)
        x = self.L2(x)
        x = self.relu1(x)
        x = self.L3(x)
        return x


if __name__ == '__main__':
    net = BLenet5(2)
    a = torch.randn(50, 1, 32, 32)
    result = net(a)
    print(result.shape)


import torch

def get_device(sout=False):
    lst = []
    # 优先使用NPU
    try:
        import torch_npu
        lst.append(torch.device("npu"))
    except ImportError:
        pass

    # 其次使用GPU
    if torch.cuda.is_available():
        lst.append(torch.device("cuda"))

    # 最后使用CPU
    lst.append(torch.device("cpu"))

    lst_str = [str(i) for i in lst]
    if sout:
        print(f"可用设备:{lst_str}, 使用{lst_str[0]}")
    return lst[0]

if __name__ == '__main__':
    result = get_device()
    print(result)


import time

from ..drawer.terminal import BColor

COLOR_DICT = {
    "default": BColor.RESET,

    "black": BColor.BLACK,
    "red": BColor.RED,
    "green": BColor.GREEN,
    "yellow": BColor.YELLOW,
    "blue": BColor.BLUE,
    "purple": BColor.PURPLE,
    "cyan": BColor.CYAN,
    "silver": BColor.SILVER
}

class BWriter:
    def __init__(self, file=None, ifTime=False, color='default'):
        '''
        :param file: 日志保存路径
        :param ifTime: 是否输出时间
        '''
        self.file = file
        self.ifTime = ifTime

        self.toWant_file = False
        self.toWant_cmd = False

        assert color in COLOR_DICT, f"color参数错误，请输入{list(COLOR_DICT.keys())}"
        self.color = color

        self.f = None
        if self.file:
            self.f = open(self.file, "a", encoding="utf-8")

    def setFile(self, file, ifTime=False):
        '''

        :param file: 设置log路径
        :param ifTime:
        :return:
        '''
        if self.f:
            self.f.close()
        self.file = file
        self.ifTime = ifTime
        self.f = open(self.file, "a", encoding="utf-8")

    def clearFile(self):
        '''
        清空内容
        :return:
        '''
        assert self.f is not None, "请先调用setFile方法"
        self.f.close()
        self.f = open(self.file, 'w', encoding="utf-8")

    def closeFile(self):
        '''
        关闭log
        :return:
        '''
        if self.f:
            self.f.close()
            self.f = None

    def toCmd(self, string, color=None):
        '''
        打印到terminal
        :param string:
        :param color:
        :return:
        '''
        # 检查color是否在字典中
        if color is None:
            print(COLOR_DICT.get(self.color) + string + BColor.RESET)
        else:
            assert color in COLOR_DICT, f"color参数错误，请输入{COLOR_DICT.keys()}"
            print(COLOR_DICT.get(color) + string + BColor.RESET)

    def toFile(self, string, ifTime=False):
        '''
        写入到文件内
        '''
        assert self.f is not None, "请先调用setFile方法"

        if ifTime or self.ifTime:
            t = time.strftime("%Y-%m-%d %H:%M:%S ##### ", time.localtime())
            self.f.write(t)

        self.f.write(string)
        self.f.write("\n")
        self.f.flush()

    def toBoth(self, string, ifTime=False, color=None):
        '''
        同时写入到文件和terminal
        :param string:
        :param color:
        :return:
        '''
        self.toFile(string, ifTime)
        self.toCmd(string, color)

    def setWant(self, toCmd=False, toFile=False):
        '''
        toWant的全局选项设置
        '''
        self.toWant_cmd = toCmd
        self.toWant_file = toFile

    def toWant(self, string, ifTime=False, color=None):
        '''
        使用前先调用setWant方法
        '''
        if self.toWant_cmd and self.toWant_file:
            self.toBoth(string, ifTime, color)
        elif self.toWant_cmd:
            self.toCmd(string, color)
        elif self.toWant_file:
            self.toFile(string, ifTime)
        else:
            raise Exception("请先调用setWant方法, 设置toCmd或toFile为True")



class BStopByLossDelta:
    def __init__(self, rounds, delta=0.002):
        '''
        连续rounds次, |before_loss - now_loss| <= delta, 则停止训练
        '''
        self.rounds = rounds
        self.delta = delta
        self.before_loss = float('inf')
        self.cnt = 0
    def __call__(self, train_loss):
        if -self.delta <= (train_loss - self.before_loss) <= self.delta:
            self.cnt += 1
            self.before_loss = train_loss
        elif self.cnt > 0:
            self.cnt -= 1
            self.before_loss = train_loss
        else:
            self.before_loss = train_loss

        if self.cnt > self.rounds:
            return True
        return False


import torch
from torch import nn

from ..trainers import BTrainer
from ..utils import get_device

class testNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.layer0 = nn.Sequential(
            nn.Linear(1*28*28, 64),
            nn.ReLU(),
            nn.Linear(64, 10)
        )
    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = self.layer0(x)
        return x

def get_dataloader(batch_size):
    from torch.utils.data import TensorDataset
    from torch.utils.data import random_split, DataLoader
    # 随机生成数据张量和标签张量
    data = torch.randn(6000, 1, 28, 28)
    labels = torch.randint(0, 10, (6000,)) # 标签为0~9, shape=(6000,)
    # 创建TensorDataset对象
    datasets = TensorDataset(data, labels)
    # 随机划分训练集和验证集
    train_datasets, val_datasets = random_split(datasets, [5500, 500])
    # 创建DataLoader对象
    train_dataloader = DataLoader(train_datasets, batch_size=batch_size, shuffle=True)
    val_dataloader = DataLoader(val_datasets, batch_size=8, shuffle=False)
    return train_dataloader, val_dataloader

def test():
    print(f"{'='*10} test.py start {'='*10}")

    #### 超参数 ####
    lr = 1e-4
    batch_size = 64
    epochs = 5
    device = get_device(sout=True)

    #### 数据集 ####
    train_dataloader, val_dataloader = get_dataloader(batch_size)
    print(f"{'='*10} TensorDataset 创建完成 {'='*10}")

    #### 模型 ####
    net = testNet()
    net.to(device)
    criterion = nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(net.parameters(), lr=lr)
    #### Trainer ####
    myTrainer = BTrainer(net, train_dataloader, val_dataloader, optimizer, criterion, device,
                         isParallel=True)
    #### 训练 ####
    myTrainer.train_eval_s(epochs)
    print(f"{'='*10} test.py success {'='*10}")

if __name__ == '__main__':
    test()